/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
#ifndef SHELL_TRAY_ICON_PRIVATE_H
#define SHELL_TRAY_ICON_PRIVATE_H

#include "shell-tray-icon.h"

#include "na-tray-child.h"

ClutterActor * shell_tray_icon_new (NaTrayChild *tray_child);

#endif /* SHELL_TRAY_ICON_PRIVATE_H */
