/* Switches */

// these are equal to the size of the SVG assets
$switch_height: 26px;
$switch_width: 48px;

.toggle-switch {
  color: $fg_color;
  height: $switch_height;
  width: $switch_width;
  background-size: contain;
  background-image: if($variant == 'light', url("resource:///org/gnome/shell/theme/toggle-off-light.svg"),url("resource:///org/gnome/shell/theme/toggle-off.svg"));
  &:checked { 
    background-image: if($variant == 'light', url("resource:///org/gnome/shell/theme/toggle-on-light.svg"),url("resource:///org/gnome/shell/theme/toggle-on.svg"));
  }
}
