.quick-settings {
  padding: $base_padding*3;
  border-radius: $modal_radius*2.25;

  .icon-button, .button {
    padding: $base_padding*1.75;
  }
}

.quick-settings-grid {
  spacing-rows: $base_padding*2;
  spacing-columns: $base_padding*2;
}

.quick-toggle, .quick-menu-toggle {
  border-radius: 99px;
  min-width: 12em;
  max-width: 12em;
  min-height: 48px;
  border:none;
}

.quick-toggle {
  &:checked { @include button(default, $c:$selected_bg_color); }

  & > StBoxLayout { spacing: $base_padding*1.5; }

  /* Move padding into the box; this is to allow menu arrows
     to extend to the border */
  &.button { padding: 0; }
  & > StBoxLayout { padding: 0 $base_padding*2; }
  &:ltr > StBoxLayout { padding-left: $base_padding*2.5; }
  &:rtl > StBoxLayout { padding-right: $base_padding*2.5; }

  .quick-toggle-title { font-weight: bold; }

  & StBoxLayout > .quick-toggle-subtitle {
    font-weight: normal;
    font-size: 12px;
  }

  .quick-toggle-icon { icon-size: $base_icon_size; }
}

.quick-menu-toggle {
  & .quick-toggle {
    min-width: auto;
    max-width: auto;

    &:ltr { border-radius: 99px 0 0 99px; }
    &:ltr > StBoxLayout { padding-right: $base_padding*1.5; }
    &:rtl { border-radius: 0 99px 99px 0; }
    &:rtr > StBoxLayout { padding-left: $base_padding*1.5; }

    &:ltr:last-child { border-radius: 99px; }
    &:rtl:last-child { border-radius: 99px; }
  }

  & .quick-toggle-arrow {
    padding: $base_padding $base_padding*1.75;
    border-width: 0;
    border-color: transparentize($fg_color, .75);

    &:checked {
      @include button(default, $c:$selected_bg_color);
      border-color: $selected_borders_color;
    }

    &:ltr {
      border-radius: 0 99px 99px 0;
      border-left-width: 1px;
    }
    &:rtl {
      border-radius: 99px 0 0 99px;
      border-right-width: 1px;
    }
  }
}

.quick-slider {
  & > StBoxLayout { spacing: $base_padding; }

  .icon-button { padding: $base_padding; }

  .slider-bin {
    &:focus {@include button(focus);}
    min-height: 16px; // slider size
    padding: $base_padding;
    border-radius: 99px;
  }
}

.quick-toggle-menu {
  @include card;

  // override some card styles
  background-color: $bg_color !important;
  color: $fg_color !important;

  border-radius: $base_border_radius*3;
  padding: $base_padding*2;
  margin: $base_padding*2 $base_padding*3 0;

  .popup-menu-item > StIcon { -st-icon-style: symbolic; }

  & .header {
    spacing-rows: 0.5 * $base_padding;
    spacing-columns: $base_padding*2;
    padding-bottom: 2 * $base_padding;

    & .icon {
      icon-size: $base_icon_size*1.5; // a non-standard symbolic size but ok
      border-radius: 999px;
      padding: 1.5 * $base_padding;
      background-color: transparentize($fg_color, 0.8);

      &.active {
        background-color: $selected_bg_color;
        color: $selected_fg_color;
      }
    }

    & .title {
      @extend %title_3;
    }

    & .subtitle {
      @extend %caption_heading;
    }
  }
}

.quick-toggle-menu-container {
}

.quick-settings-system-item {
  & > StBoxLayout { spacing: 2 * $base_padding; }

  & .power-item {
    min-height: 0;
    min-width: 0;

    &:insensitive {
      @include button(normal);
      background-color: transparent;
    }
  }
}

.nm-network-item {
  .wireless-secure-icon { icon-size: 0.5 * $base_icon_size; }
}

.bt-device-item {
  .popup-menu-icon { -st-icon-style: symbolic; }
}

.bt-menu-placeholder.popup-menu-item {
  @extend %title_4;
  text-align: center;

  padding: 2em 4em;
}

.device-subtitle { color: transparentize($fg_color, 0.5); }

.keyboard-brightness-item .slider {
  min-height: 16px;
}

.keyboard-brightness-level {
  spacing: $base_padding;

  .button:checked { @include button(default, $c:$selected_bg_color); }
}

// background apps

.background-apps-quick-toggle {
  min-height: 40px;
  background-color: transparent;

  & StIcon { icon-size: $base_icon_size !important; }
}

.background-app-item {
  & .title { @extend %heading; }
  & .subtitle { @extend %caption; }
  & .popup-menu-icon {
    icon-size: $large_icon_size !important;
    -st-icon-style: regular !important;
  }
  & .close-button {
    @extend .icon-button;
    padding: $base_padding;
  }
  & .spinner {
    padding: $base_padding;
  }

  &.popup-inactive-menu-item { color: $fg_color; }
}
