/* Dash */

// uses system colors
$dash_background_color: $system_overlay_bg_color;

$dash_placeholder_size: 32px;
$dash_padding: $base_padding*2; // 12px
$dash_border_radius: $modal_radius + $dash_padding;

// container for the dash
#dash {
  @extend %caption;
  margin-top: $dash_padding;

  .placeholder {
    // background-image: url("resource:///org/gnome/shell/theme/dash-placeholder.svg");
    background-image:none;
    background-size: contain;
    height: $dash_placeholder_size;
  }

  .empty-dash-drop-target {
    width: $dash_placeholder_size;
    height: $dash_placeholder_size;
  }

  // Running app indicator (also shown in app grid)
  .app-well-app-running-dot {
    margin-bottom: 14px; // hardcoded
  }
}

// background of the dash behind app icons
.dash-background {
  background-color: $dash_background_color;
  border-radius: $dash_border_radius;
  padding: $dash_padding;
  spacing: $base_padding;
  margin-bottom: $dash_padding;

  @if $is_highcontrast {
    border: 2px solid $hc_inset_color;
    box-shadow:none;
  }
}

// items on the dash
.dash-item-container {
  > * {margin: 0 2px;}
  &:ltr{&:first-child {margin-left: 0;}}
  &:rtl{&:first-child {margin-right: 0;}}

  // each app item on the dash
  .app-well-app .overview-icon {
    padding: $base_padding;
    border-radius: $base_border_radius*2;
  }

  // show apps button
  .show-apps { @include overview_icon($system_fg_color);}

  .show-apps, .app-well-app {
    padding-bottom: $dash_padding;
    color: $system_fg_color;
  }
}

// separator between favourites and running apps
.dash-separator {
  width: 1px;
  margin: 0 $base_margin;
  background-color: $system_borders_color;
  margin-bottom: $dash_padding;
  @if $is_highcontrast {
    width: 2px;
  }
}

// OSD Tooltip
.dash-label {
  @extend %tooltip;
  -y-offset: $base_margin * 2; // distance from the dash edge
}
