/* Entries */

StEntry {
  @extend %entry;

  StIcon.capslock-warning {
    icon-size: $base_icon_size;
    warning-color: $warning_color;
    padding: 0 4px;
  }

  StIcon.peek-password {
    icon-size: $base_icon_size;
    padding: 0 4px;
  }

  StLabel.hint-text {
    margin-left: 2px;
  }
}
